//
// Created by philipp on 04.04.18.
//

#include "../include/PlatoonProtocolLib/VehicleFacade.h"
#include "../include/PlatoonProtocolLib/LeaderVehicle.h"
#include "../include/PlatoonProtocolLib/FollowerVehicle.h"

namespace platoonProtocol
{

VehicleFacade::VehicleFacade(Vehicle::Role role, networking::Networking & net, VehicleId vehicleId) :
  net(net), vehicleId(vehicleId)
{
    reset(role);
}

void VehicleFacade::reset(Vehicle::Role role)
{
    std::lock_guard<std::mutex> lock(mutex);
    if (role == Vehicle::Role::LEADER)
        vehicle = LeaderVehicle::create(net, vehicleId);
    else
    {
        vehicle = FollowerVehicle::create(net, vehicleId);
        std::dynamic_pointer_cast<FollowerVehicle>(vehicle)->setOnPlatoonConfigUpdatedCallback(callback);
    }
    vehicle->setOnLeavingPlatoonCallback(callback);
    vehicle->setOnRunningPlatoonCallback(callback);
    vehicle->addVehicleEndpoints(vehicleEndpoints);
}

Vehicle::Ptr VehicleFacade::get()
{
    std::lock_guard<std::mutex> lock(mutex);
    return vehicle;
}

bool VehicleFacade::isIdle()
{
    std::lock_guard<std::mutex> lock(mutex);
    return vehicle->isIdle();
}

bool VehicleFacade::isPlatoonRunning()
{
    std::lock_guard<std::mutex> lock(mutex);
    return vehicle->isPlatoonRunning();
}

void VehicleFacade::createPlatoon()
{
    std::lock_guard<std::mutex> lock(mutex);
    vehicle->createPlatoon();
}

void VehicleFacade::leavePlatoon()
{
    std::lock_guard<std::mutex> lock(mutex);
    vehicle->leavePlatoon();
}

void VehicleFacade::stop()
{
    std::lock_guard<std::mutex> lock(mutex);
    vehicle->stop();
}

void VehicleFacade::addVehicleEndpoints(const Vehicle::VehicleEndpoints & endpoints)
{
    std::lock_guard<std::mutex> lock(mutex);
    for (const auto & endpoint : endpoints)
        vehicleEndpoints.push_back(endpoint);
    vehicle->addVehicleEndpoints(endpoints);
}

VehicleEndpoint VehicleFacade::getOwnEndpoint()
{
    std::lock_guard<std::mutex> lock(mutex);
    return vehicle->getOwnEndpoint();
}

Vehicle::Role VehicleFacade::getRole()
{
    std::lock_guard<std::mutex> lock(mutex);
    return vehicle->getRole();
}

void VehicleFacade::setCallback(const Vehicle::Callback & callback)
{
    std::lock_guard<std::mutex> lock(mutex);
    this->callback = callback;
    vehicle->setOnLeavingPlatoonCallback(callback);
    vehicle->setOnRunningPlatoonCallback(callback);
    if (vehicle->getRole() == Vehicle::Role::FOLLOWER)
        std::dynamic_pointer_cast<FollowerVehicle>(vehicle)->setOnPlatoonConfigUpdatedCallback(callback);
}

std::size_t VehicleFacade::getNumFollowers()
{
    std::lock_guard<std::mutex> lock(mutex);
    return std::dynamic_pointer_cast<LeaderVehicle>(vehicle)->getNumFollowers();
}

void VehicleFacade::setPlatoonConfig(const PlatoonConfig & platoonConfig)
{
    std::lock_guard<std::mutex> lock(mutex);
    std::dynamic_pointer_cast<LeaderVehicle>(vehicle)->setPlatoonConfig(platoonConfig);
}

void VehicleFacade::setPlatoonSpeed(PlatoonSpeed platoonSpeed)
{
    std::lock_guard<std::mutex> lock(mutex);
    std::dynamic_pointer_cast<LeaderVehicle>(vehicle)->setPlatoonSpeed(platoonSpeed);
}

void VehicleFacade::setInnerPlatoonDistance(InnerPlatoonDistance innerPlatoonDistance)
{
    std::lock_guard<std::mutex> lock(mutex);
    std::dynamic_pointer_cast<LeaderVehicle>(vehicle)->setInnerPlatoonDistance(innerPlatoonDistance);
}

networking::time::TimedValue<PlatoonConfig> VehicleFacade::getPlatoonConfig()
{
    std::lock_guard<std::mutex> lock(mutex);
    return std::dynamic_pointer_cast<FollowerVehicle>(vehicle)->getPlatoonConfig();
}

}
